"""Utility functions to abstract Python version differences."""

import sys


class UnicodeMixin(object):
  """Mixin class to handle defining the proper __str__/__unicode__
  methods in Python 2 or 3."""

  if sys.version_info[0] >= 3: # Python 3
      def __str__(self):
          return self.__unicode__()
  else:  # Python 2
      def __str__(self):
          return self.__unicode__().encode('utf8')


def _encode_utf8(unicode_string):
  if sys.version_info[0] >= 3: # Python 3
      return bytes(unicode_string, 'utf-8')
  else:
      return unicode_string.encode('utf-8')


def _decode_utf8(bytes_obj):
  if sys.version_info[0] >= 3: # Python 3
      return str(bytes_obj, 'utf-8')
  else:
      return unicode(bytes_obj, 'utf-8')




from cffi import FFI

ffi = FFI()

ffi.cdef("""
// lber type definitions
typedef ... BerElement;
// Hack because cffi does not support #include in cdef.
// These types are usually declared in lber_types.h
typedef unsigned long ber_len_t;

typedef struct berval {
    ber_len_t bv_len;
    char *bv_val;
} BerValue, *BerVarray;


// libldap type definitions
typedef ... LDAP;
typedef ... LDAPMessage;
typedef ... LDAPControl;

typedef struct ldapmod {
    int mod_op;
    char *mod_type;
    union {
        char **modv_strvals;
        struct berval **modv_bvals;
    } mod_vals;
} LDAPMod;

#define LDAP_VERSION3 ...

#define LDAP_OPT_PROTOCOL_VERSION ...
#define LDAP_OPT_X_TLS_REQUIRE_CERT ...
#define LDAP_OPT_X_TLS_CACERTFILE ...
#define LDAP_OPT_X_TLS_NEWCTX ...
#define LDAP_OPT_X_TLS_NEVER ...
#define LDAP_OPT_X_TLS_HARD ...
#define LDAP_OPT_X_TLS_DEMAND ...
#define LDAP_OPT_X_TLS_ALLOW ...
#define LDAP_OPT_X_TLS_TRY ...
#define LDAP_OPT_TIMELIMIT ...

static char *const LDAP_ALL_USER_ATTRIBUTES;
static char *const LDAP_ALL_OPERATIONAL_ATTRIBUTES;
static char *const LDAP_NO_ATTRS;

#define LDAP_NO_LIMIT ...
#define LDAP_MOD_ADD ...
#define LDAP_MOD_DELETE ...
#define LDAP_MOD_REPLACE ...
#define LDAP_MOD_BVALUES ...
#define LDAP_SCOPE_BASE ...
#define LDAP_SCOPE_ONELEVEL ...
#define LDAP_SCOPE_SUBTREE ...
#define LDAP_SUCCESS ...
#define LDAP_NO_SUCH_OBJECT ...
#define LDAP_INVALID_CREDENTIALS ...
#define LDAP_SERVER_DOWN ...

// Function declarations
int ldap_initialize(LDAP **ldp, char *uri);
int ldap_set_option(LDAP *ld, int option, const void *invalue);
int ldap_simple_bind_s(LDAP *ld, const char *who, const char *passwd);
int ldap_search_ext_s(
       LDAP *ld,
       char *base,
       int scope,
       char *filter,
       char *attrs[],
       int attrsonly,
       LDAPControl **serverctrls,
       LDAPControl **clientctrls,
       struct timeval *timeout,
       int sizelimit,
       LDAPMessage **res);

// From ldap_next_entry(3)
int ldap_count_entries( LDAP *ld, LDAPMessage *result );
LDAPMessage *ldap_first_entry( LDAP *ld, LDAPMessage *result );
LDAPMessage *ldap_next_entry( LDAP *ld, LDAPMessage *entry );

// From ldap_get_values(3)
struct berval **ldap_get_values_len(LDAP *ld, LDAPMessage *entry, char *attr);
int ldap_count_values_len(struct berval **vals);

// From ldap_get_dn(3)
char *ldap_get_dn( LDAP *ld, LDAPMessage *entry );

// From ldap_first_attribute(3)
char *ldap_first_attribute( LDAP *ld, LDAPMessage *entry, BerElement **berptr );
char *ldap_next_attribute( LDAP *ld, LDAPMessage *entry, BerElement *ber );

// From ldap_add_ext(3)
int ldap_add_ext_s(
       LDAP *ld,
       const char *dn,
       LDAPMod **attrs,
       LDAPControl **sctrls,
       LDAPControl **cctrls );

// From ldap_modify_ext(3)
int ldap_modify_ext_s(
              LDAP *ld,
              char *dn,
              LDAPMod *mods[],
              LDAPControl **sctrls,
              LDAPControl **cctrls );

// From ldap_delete_s(3)
int ldap_delete_s(LDAP *ld, char *dn);

// From ldap_rename_s(3)
int ldap_rename_s(
        LDAP *ld,
        const char *dn,
        const char *newrdn,
        const char *newparent,
        int deleteoldrdn,
        LDAPControl *sctrls[],
        LDAPControl *cctrls[]);

// From ldap_err2string(3)
char *ldap_err2string( int err );

// From ldap_msgfree(3)
int ldap_msgfree( LDAPMessage *msg );

// From ldap_memfree(3)
void ldap_memfree(void *p);

int ldap_passwd_s(
        LDAP *ld,
        struct berval        *user,
        struct berval        *oldpw,
        struct berval        *newpw,
        struct berval *newpasswd,
        LDAPControl **sctrls,
        LDAPControl **cctrls );

// From lber-types(3)
struct berval *ber_bvstr(const char *str);
void ber_bvfree(struct berval *bv);

// From ber_free(3)
void ber_free(BerElement *ber, int freebuf);

""")

libldap = ffi.verify(
"""
// Required for ldap_bind_simple
#define LDAP_DEPRECATED 1

#include <ldap.h>
#include <lber.h>
""", libraries=[str("ldap"), str("lber")])



# -*- coding: utf-8 -*-
"""A CFFI/libldap-based LDAP connection."""

from __future__ import unicode_literals
from __future__ import print_function

import copy
import sys

from ldapom.cdef import libldap, ffi
from ldapom import attribute
from ldapom import compat
from ldapom.entry import LDAPEntry
from ldapom import error

LDAP_SCOPE_BASE = libldap.LDAP_SCOPE_BASE
LDAP_SCOPE_SUBTREE = libldap.LDAP_SCOPE_SUBTREE
LDAP_SCOPE_ONELEVEL = libldap.LDAP_SCOPE_ONELEVEL

LDAP_OPT_X_TLS_NEVER = libldap.LDAP_OPT_X_TLS_NEVER
LDAP_OPT_X_TLS_HARD = libldap.LDAP_OPT_X_TLS_HARD
LDAP_OPT_X_TLS_DEMAND = libldap.LDAP_OPT_X_TLS_DEMAND
LDAP_OPT_X_TLS_ALLOW = libldap.LDAP_OPT_X_TLS_ALLOW
LDAP_OPT_X_TLS_TRY = libldap.LDAP_OPT_X_TLS_TRY


def handle_ldap_error(err):
    """Given an LDAP error code, raise an error if needed.

    :param err: The error code returned by the library.
    :type err: int
    """
    if err == libldap.LDAP_SUCCESS:
        return

    error_string = compat._decode_utf8(ffi.string(libldap.ldap_err2string(err)))
    if err == libldap.LDAP_NO_SUCH_OBJECT:
        raise error.LDAPNoSuchObjectError(error_string)
    elif err == libldap.LDAP_INVALID_CREDENTIALS:
        raise error.LDAPInvalidCredentialsError(error_string)
    elif err == libldap.LDAP_SERVER_DOWN:
        raise error.LDAPServerDownError(error_string)
    else:
        raise error.LDAPError(error_string)


def _retry_reconnect(func):
    def func_wrapper(lc, *args, **kwargs):
        retry_count = 0
        while True:
            try:
                return func(lc, *args, **kwargs)
                break
            except error.LDAPServerDownError:
                if retry_count >= lc._max_retry_reconnect:
                    raise

                retry_count += 1
                try:
                    lc._connect()
                except error.LDAPServerDownError:
                    pass


    return func_wrapper

def _retry_reconnect_generator(func):
    def generator_wrapper(lc, *args, **kwargs):
        retry_count = 0
        while True:
            try:
                g = func(lc, *args, **kwargs)
                for v in g:
                    yield v
                break
            except error.LDAPServerDownError:
                if retry_count >= lc._max_retry_reconnect:
                    raise

                retry_count += 1
                try:
                    lc._connect()
                except error.LDAPServerDownError:
                    pass
    return generator_wrapper

class LDAPConnection(object):
    """Connection to an LDAP server."""

    def __init__(self, uri, base, bind_dn, bind_password,
            cacertfile=None, require_cert=LDAP_OPT_X_TLS_NEVER,
            timelimit=30, max_retry_reconnect=5,
            schema_base="cn=subschema", enable_attribute_type_mapping=True,
            retrieve_operational_attributes=False):
        """
        :param uri: URI of the server to connect to.
        :param base: Base DN for LDAP operations.
        :param login: DN to bind with.
        :param password: Password to bind with.
        :param cacertfile: If using SSL/TLS this is certificate of the server
        :param timelimit: Defines the time limit after which a search
            operation should be terminated by the server
        :param schema_base: base DN for the schema description.
        :param enable_attribute_type_mapping: Whether to enable the mapping of LDAP attribute types
            to corresponding Python types. Requires the schema to be fetched when connecting. If
            disabled, all attributes will be treated as a multi-value string attribute.
        """
        self._base = base
        self._uri = uri
        self._bind_dn = bind_dn
        self._bind_password = bind_password
        self._cacertfile = cacertfile
        self._require_cert = require_cert
        self._max_retry_reconnect = max_retry_reconnect
        self._timelimit = timelimit
        self._schema_base = schema_base
        self._enable_attribute_type_mapping = enable_attribute_type_mapping

        self._connect()

    def _connect(self):
        ld_p = ffi.new("LDAP **")
        err = libldap.ldap_initialize(ld_p, compat._encode_utf8(self._uri))
        handle_ldap_error(err)
        self._ld = ld_p[0]

        version_p = ffi.new("int *")
        version_p[0] = libldap.LDAP_VERSION3
        libldap.ldap_set_option(self._ld, libldap.LDAP_OPT_PROTOCOL_VERSION, version_p)

        if self._cacertfile is not None:
            libldap.ldap_set_option(self._ld, libldap.LDAP_OPT_X_TLS_CACERTFILE,
                    compat._encode_utf8(self._cacertfile))

        require_cert_p = ffi.new("int *")
        require_cert_p[0] = self._require_cert
        libldap.ldap_set_option(self._ld, libldap.LDAP_OPT_X_TLS_REQUIRE_CERT,
                require_cert_p);

        # For TLS options to take effect, a context refresh seems to be needed.
        newctx_p = ffi.new("int *")
        newctx_p[0] = 0
        libldap.ldap_set_option(self._ld, libldap.LDAP_OPT_X_TLS_NEWCTX, newctx_p)

        timelimit_p = ffi.new("int *")
        timelimit_p[0] = self._timelimit
        libldap.ldap_set_option(self._ld, libldap.LDAP_OPT_TIMELIMIT, timelimit_p)

        err = libldap.ldap_simple_bind_s(self._ld,
                compat._encode_utf8(self._bind_dn),
                compat._encode_utf8(self._bind_password))
        handle_ldap_error(err)

        if self._enable_attribute_type_mapping:
            self._fetch_attribute_types()

    def _fetch_attribute_types(self):
        attribute_type_definitions = attribute.DEFAULT_ATTRIBUTE_TYPES
        result = list(
                self._raw_search(base=self._schema_base,
                    scope=libldap.LDAP_SCOPE_BASE,
                    search_filter="(objectClass=*)",
                    retrieve_attributes=["attributeTypes"]))
        # Decode the type definitions returned to strings
        attribute_type_definitions += map(compat._decode_utf8,
                result[0][1][compat._encode_utf8("attributeTypes")])

        self._attribute_types_by_name = attribute.build_attribute_types(
                attribute_type_definitions)

    def get_attribute_type(self, name):
        """Get the Python type to represent an attribute.

        :param name: The name of the attribute to look up the type for.
        :type name: str
        :rtype: a class object, a subclass of ``LDAPAttributeBase``.
        """
        if self._enable_attribute_type_mapping:
            if name in self._attribute_types_by_name:
                return self._attribute_types_by_name[name]
            else:
                raise error.LDAPAttributeNameNotFoundError(
                        'Attribute type "{}" not found.'.format(name))
        else:
            # Use a multi-value string attribute as the default
            base_classes = [attribute.MultiValueAttributeMixin,
                            attribute.BytesAttributeMixin,
                            attribute.LDAPAttributeBase]
            if sys.version_info[0] >= 3: # Python 3
                return type("LDAPAttribute", tuple(base_classes), {})
            else:
                return type(bytes("LDAPAttribute"), tuple(base_classes), {})

    @_retry_reconnect
    def can_bind(self, bind_dn, bind_password):
        """Try to bind with the given credentials.

        :param bind_dn: DN to bind with.
        :type bind_dn: str
        :param bind_password: Password to bind with.
        :type bind_password: str
        :rtype: boolean
        """
        try:
            self.__class__(self._uri, self._base, bind_dn, bind_password,
                    self._cacertfile)
        except error.LDAPInvalidCredentialsError:
            return False
        return True

    def _raw_search(self, search_filter=None, retrieve_attributes=None,
            base=None, scope=libldap.LDAP_SCOPE_SUBTREE,
            retrieve_operational_attributes=False):
        """
        Raw wrapper around OpenLDAP ldap_search_ext_s.

        :param search_filter: Filter expression to use. OpenLDAP default used
            if None is given.
        :type search_filter: List of str
        :param retrieve_attributes: List of attributes to retrieve. If None is
            given, all user attributes are retrieved.
        :type retrieve_attributes: List of str
        :param base: Search base for the query.
        :type base: str
        :param scope: The search scope in the LDAP tree
        :param retrieve_operational_attributes: Retrieve operational attributes of entries in
            addition to user attributes if retrieve_attributes is not set.
        """
        search_result_p = ffi.new("LDAPMessage **")

        # Keep around references to pointers to owned memory with data that is
        # still needed.
        prevent_garbage_collection = []

        if retrieve_attributes is None:
            retrieve_attributes = [
                compat._decode_utf8(ffi.string(libldap.LDAP_ALL_USER_ATTRIBUTES))]
            if retrieve_operational_attributes:
                retrieve_attributes.append(
                    compat._decode_utf8(ffi.string(libldap.LDAP_ALL_OPERATIONAL_ATTRIBUTES)))

        attrs_p = ffi.new("char*[{}]".format(len(retrieve_attributes) + 1))
        for i, a in enumerate(retrieve_attributes):
            attr_p = ffi.new("char[]", compat._encode_utf8(a))
            prevent_garbage_collection.append(attr_p)
            attrs_p[i] = attr_p
        attrs_p[len(retrieve_attributes)] = ffi.NULL

        err = libldap.ldap_search_ext_s(
                self._ld,
                compat._encode_utf8(base or self._base),
                scope,
                (compat._encode_utf8(search_filter)
                    if search_filter is not None else ffi.NULL),
                attrs_p,
                0,
                ffi.NULL, ffi.NULL,
                ffi.NULL, # TODO: Implement timeouts
                0,#libldap.LDAP_NO_LIMIT,
                search_result_p)
        handle_ldap_error(err)
        search_result = search_result_p[0]

        current_entry = libldap.ldap_first_entry(self._ld, search_result)
        while current_entry != ffi.NULL:
            dn = ffi.string(libldap.ldap_get_dn(self._ld, current_entry))
            attribute_dict = {}

            ber_p = ffi.new("BerElement **")
            current_attribute = libldap.ldap_first_attribute(self._ld,
                    current_entry, ber_p)
            while current_attribute != ffi.NULL:
                current_attribute_str = ffi.string(current_attribute)
                attribute_dict[current_attribute_str] = []

                values_p = libldap.ldap_get_values_len(self._ld, current_entry,
                        current_attribute)
                for i in range(0, libldap.ldap_count_values_len(values_p)):
                    val = ffi.buffer(values_p[i].bv_val, values_p[i].bv_len)[:]
                    attribute_dict[current_attribute_str].append(val)

                libldap.ldap_memfree(current_attribute)
                current_attribute = libldap.ldap_next_attribute(self._ld,
                        current_entry, ber_p[0])
            libldap.ber_free(ber_p[0], 0)

            yield (dn, attribute_dict)
            current_entry = libldap.ldap_next_entry(self._ld, current_entry)

        libldap.ldap_msgfree(search_result)

    @_retry_reconnect_generator
    def search(self, *args, **kwargs):
        """Perform an LDAP search operation."""
        return self._search(*args, **kwargs)

    def _search(self, search_filter=None, retrieve_attributes=None,
            base=None, scope=libldap.LDAP_SCOPE_SUBTREE,
            retrieve_operational_attributes=False):
        """Search without retry_reconnect."""
        try:
            raw_search_result = self._raw_search(search_filter=search_filter,
                                                 retrieve_attributes=retrieve_attributes,
                                                 base=base,scope=scope,
                                                 retrieve_operational_attributes=retrieve_operational_attributes)
            for dn, attributes_dict in raw_search_result:
                entry = LDAPEntry(self, compat._decode_utf8(dn),
                                  retrieve_attributes=retrieve_attributes,
                                  retrieve_operational_attributes=retrieve_operational_attributes)
                entry._attributes = set()
                for name, value in attributes_dict.items():
                    # TODO: Create the right type of LDAPAttribute here
                    attribute_type = self.get_attribute_type(
                            compat._decode_utf8(name))
                    attribute = attribute_type(compat._decode_utf8(name))
                    attribute._set_ldap_values(value)
                    entry._attributes.add(attribute)
                entry._fetched_attributes = copy.deepcopy(entry._attributes)
                yield entry
        except error.LDAPNoSuchObjectError:
            # If the search returned without results, "return" an empty generator.
            return

    def get_entry(self, *args, **kwargs):
        """Get an LDAPEntry object associated with this connection."""
        return LDAPEntry(self, *args, **kwargs)

    @_retry_reconnect
    def delete(self, entry, recursive=False):
        """Delete an entry on the LDAP server.

        :param entry: The entry to delete.
        :type entry: ldapom.LDAPEntry
        :param recursive: If subentries should be deleted recursively.
        :type recursive: bool
        """
        if recursive:
            entries_to_delete = self._connection._search(
                    base=entry.dn,
                    scope=LDAP_SCOPE_ONELEVEL)
            for entry in entries_to_delete:
                entry.delete(recursive=True)
        err = libldap.ldap_delete_s(self._ld,
                compat._encode_utf8(entry.dn))
        handle_ldap_error(err)

    @_retry_reconnect
    def rename(self, entry, new_dn):
        """Rename an entry on the LDAP server.

        :param entry: The entry to rename.
        :type entry: ldapom.LDAPEntry
        :param new_dn: The DN that the entry should have after the rename.
        :type new_dn: str
        """
        new_rdn, new_parent_dn = new_dn.split(",", 1)
        if new_parent_dn == entry.parent_dn:
            new_parent_dn = None

        err = libldap.ldap_rename_s(self._ld,
                compat._encode_utf8(entry.dn),
                compat._encode_utf8(new_rdn),
                (compat._encode_utf8(new_parent_dn)
                    if new_parent_dn is not None else ffi.NULL),
                1, # Delete old RDN
                ffi.NULL, ffi.NULL)
        handle_ldap_error(err)

        entry._dn = new_dn

    @_retry_reconnect
    def exists(self, entry):
        """Checks if a the given entry exists on the LDAP server.

        :param entry: The entry to check the existence of.
        :type entry: ldapom.LDAPEntry
        """
        entry_search_result = list(self._search(
            base=entry.dn, scope=LDAP_SCOPE_BASE))
        return len(entry_search_result) == 1

    @_retry_reconnect
    def save(self, entry):
        """Save the given entry and its attribute values to the LDAP server.

        :param entry: The entry to save.
        :type entry: ldapom.LDAPEntry
        """
        entry_exists = entry.exists()
        # Refuse to save if attributes have not been fetched or set explicitly.
        if entry._attributes is None:
            raise error.LDAPomError("Cannot save without attributes "
                    "previously fetched or set.")

        if entry_exists:
            assert entry._fetched_attributes is not None
            changed_attributes = entry._attributes - entry._fetched_attributes
            # Deleted attributes are represented as empty attributes to the LDAP server.
            deleted_attribute_names = (frozenset(a.name for a in entry._fetched_attributes)
                    - frozenset(a.name for a in entry._attributes))
            for deleted_name in deleted_attribute_names:
                deleted_attribute_type = self.get_attribute_type(deleted_name)
                changed_attributes.add(deleted_attribute_type(deleted_name))
        else:
            # Don't try to save empty attributes as this fails if the entry does
            # not exist on the server yet.
            changed_attributes = set(filter(lambda attr: len(attr._values) > 0, entry._attributes))

        # Keep around references to pointers to owned memory with data that is
        # still needed.
        prevent_garbage_collection = []

        mods = ffi.new("LDAPMod*[{}]".format(len(changed_attributes) + 1))
        for i, attribute in enumerate(changed_attributes):
            mod = ffi.new("LDAPMod *")
            prevent_garbage_collection.append(mod)

            mod.mod_op = libldap.LDAP_MOD_REPLACE | libldap.LDAP_MOD_BVALUES

            mod_type = ffi.new("char[]", compat._encode_utf8(attribute.name))
            prevent_garbage_collection.append(mod_type)
            mod.mod_type = mod_type

            modv_bvals = ffi.new("BerValue*[{}]".format(len(attribute._values) + 1))
            prevent_garbage_collection.append(modv_bvals)
            for j, value in enumerate(attribute._get_ldap_values()):
                modv_berval = ffi.new("BerValue *")
                prevent_garbage_collection.append(modv_berval)
                modv_berval.bv_len = len(value)
                bval = ffi.new("char[]", len(value))
                prevent_garbage_collection.append(bval)
                ffi.buffer(bval)[:] = value
                modv_berval.bv_val = bval
                modv_bvals[j] = modv_berval
            modv_bvals[len(attribute._values)] = ffi.NULL
            mod.mod_vals = {"modv_bvals": modv_bvals}
            mods[i] = mod
        mods[len(changed_attributes)] = ffi.NULL

        if entry_exists:
            err = libldap.ldap_modify_ext_s(self._ld,
                    compat._encode_utf8(entry.dn),
                    mods,
                    ffi.NULL, ffi.NULL)
        else:
            err = libldap.ldap_add_ext_s(self._ld,
                    compat._encode_utf8(entry.dn),
                    mods,
                    ffi.NULL, ffi.NULL)
        handle_ldap_error(err)

        entry._fetched_attributes = copy.deepcopy(entry._attributes)

    @_retry_reconnect
    def fetch(self, entry, *args, **kwargs):
        """Fetch an entry's attributes from the LDAP server.

        :param entry: The entry to fetch.
        :type entry: ldapom.LDAPEntry
        """
        try:
            fetched_entry = next(self._search(*args, base=entry.dn,
                scope=libldap.LDAP_SCOPE_BASE, **kwargs))
            entry._attributes = fetched_entry._attributes
            entry._fetched_attributes = copy.deepcopy(entry._attributes)
        except StopIteration:
            raise error.LDAPNoSuchObjectError()

    @_retry_reconnect
    def set_password(self, entry, password):
        """Set the bind password for an entry.

        :param entry: The entry to set the password for.
        :type entry: ldapom.LDAPEntry
        :param password: The password to set.
        :type password: str
        """
        password_p = ffi.new("char[]", compat._encode_utf8(password))
        password_berval = libldap.ber_bvstr(password_p)
        entry_dn_p = ffi.new("char[]", compat._encode_utf8(entry.dn))
        entry_dn_berval = libldap.ber_bvstr(entry_dn_p)

        err = libldap.ldap_passwd_s(self._ld,
                entry_dn_berval,
                ffi.NULL,
                password_berval, password_berval,
                ffi.NULL, ffi.NULL)
        handle_ldap_error(err)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function


class LDAPomError(Exception):
    pass


class LDAPError(LDAPomError):
    pass


class LDAPNoSuchObjectError(LDAPError):
    pass


class LDAPInvalidCredentialsError(LDAPError):
    pass


class LDAPServerDownError(LDAPError):
    pass


class LDAPAttributeNameNotFoundError(LDAPomError):
    pass



# -*- coding: utf-8 -*-
"""Types for LDAP attributes."""

from __future__ import unicode_literals
from __future__ import print_function

from ldapom import compat


class LDAPEntry(compat.UnicodeMixin, object):
    """Lazy-loading LDAP entry object."""

    def __init__(self, connection, dn,
                 retrieve_attributes=None, retrieve_operational_attributes=False):
        """Creates a lazy entry object by dn.

        :param connection: The connection to use for this node.
        :type connection: LdapConnection
        :param dn: The DN for this node.
        :type dn: str
        :param retrieve_attributes: List of attributes to retrieve. If None is
            given, all user attributes are retrieved.
        :param retrieve_operational_attributes: Retrieve operational attributes of entries in
            addition to user attributes if retrieve_attributes is not set.
        """
        # Use super() method because __setattr__ is overridden.
        super(LDAPEntry, self).__setattr__('_connection', connection)
        super(LDAPEntry, self).__setattr__('_dn', dn)
        super(LDAPEntry, self).__setattr__('_attributes', None)
        super(LDAPEntry, self).__setattr__('_fetched_attributes', None)
        super(LDAPEntry, self).__setattr__('_retrieve_attributes', retrieve_attributes)
        super(LDAPEntry, self).__setattr__('_retrieve_operational_attributes', retrieve_operational_attributes)

    ## Expose dn as a ready-only property
    dn = property(lambda self: self._dn)
    rdn = property(lambda self: self.dn.split(",")[0])
    parent_dn = property(lambda self: ",".join(self.dn.split(",")[1:]))

    def fetch(self):
        """Fetch this entry's attributes from the LDAP server."""
        return self._connection.fetch(
            self, retrieve_attributes=self._retrieve_attributes,
            retrieve_operational_attributes=self._retrieve_operational_attributes)

    def _fetch_attributes_if_exists(self):
        """Fetch this entry's attributes from the LDAP server if it exists.

        If it doesn't exist, simply use an empty set as attributes.
        """
        if self._attributes is not None:
            return

        if self.exists():
            self.fetch()
        else:
            self._attributes = set()

    def exists(self):
        """Checks if this entry exists on the LDAP server."""
        return self._connection.exists(self)

    def get_attribute(self, name):
        """Get a named attribute object from the internal set of attributes.

        :param name: The name of the attribute to get.
        :type name: str
        :rtype: LDAPAttribute object or None
        """
        try:
            return [a for a in self._attributes if a.name == name][0]
        except IndexError:
            return None

    def __getattr__(self, name):
        """Get an attribute or query object class membership"""
        self._fetch_attributes_if_exists()

        if name.startswith("is_"):
            return name[3:] in self.get_attribute("objectClass").values

        attribute_type = self._connection.get_attribute_type(name)
        attribute = self.get_attribute(name)
        if attribute is not None:
            if attribute.single_value:
                return attribute.value
            else:
                return attribute.values
        else:
            if attribute_type.multi_value:
                setattr(self, name, set())
                return self.get_attribute(name).values
            else:
                raise AttributeError()

    def __setattr__(self, name, value):
        """Set an attribute value.

        If the attribute is multi-value but the passed value is not a list
        or set, the value is set as the first and only value of the set of
        values for this attribute.
        """
        # Use normal behaviour if setting an existing instance attribute.
        if name in self.__dict__:
            super(LDAPEntry, self).__setattr__(name, value)
            return

        self._fetch_attributes_if_exists()

        # Try to get existing attribute from list, if not found
        # create a new once.
        attribute = self.get_attribute(name)
        if attribute is None:
            attribute_type = self._connection.get_attribute_type(name)
            attribute = attribute_type(name)
            self._attributes.add(attribute)

        if attribute.single_value:
            attribute.value = value
        else:
            if isinstance(value, (list, set)):
                attribute.values = set(value)
            else:
                attribute.values = {value}

    def __delattr__(self, name):
        self._fetch_attributes_if_exists()
        attribute = self.get_attribute(name)
        if attribute is not None:
            self._attributes.remove(attribute)

    def __unicode__(self):
        return self.dn

    def __repr__(self):
        return self.__str__()

    def delete(self, *args, **kwargs):
        """Delete this entry on the LDAP server."""
        return self._connection.delete(self, *args, **kwargs)

    def rename(self, *args, **kwargs):
        """Rename this entry on the LDAP server."""
        return self._connection.rename(self, *args, **kwargs)

    def save(self, *args, **kwargs):
        """Save this entry and its attribute values to the LDAP server."""
        return self._connection.save(self, *args, **kwargs)

    def set_password(self, password):
        """Change the password for this LDAP entry.

        :param password: The password to set
        :type password: str
        """
        return self._connection.set_password(self, password)

    def can_bind(self, bind_password):
        """Try to bind with the given credentials.

        :param bind_password: Password to bind with.
        :type bind_password: str
        :rtype: boolean
        """
        return self._connection.can_bind(self.dn, bind_password)



# -*- coding: utf-8 -*-

import os
from setuptools import setup
from distutils.command.build import build

class cffi_build(build):
    """This is a shameful hack to ensure that cffi is present when
    we specify ext_modules. We can't do this eagerly because
    setup_requires hasn't run yet.

    Copied from https://github.com/xattr/xattr
    """
    def finalize_options(self):
        import ldapom
        self.distribution.ext_modules = [ldapom.connection.ffi.verifier.get_extension()]
        build.finalize_options(self)

# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='ldapom',
      version='0.12.0',
      description='A pythonic LDAP to Python object mapper',
      url='https://github.com/leonhandreke/ldapom',
      license='MIT',
      keywords = "ldap object mapper",
      long_description=read('README.md'),
      packages=['ldapom',],
      install_requires=['cffi'],
      setup_requires=['cffi'],
      zip_safe=False,
      cmdclass={'build': cffi_build},
     )




# -*- coding: utf-8 -*-
"""Types for dealing with different LDAP attributes."""

from __future__ import unicode_literals
from __future__ import print_function

import copy
import re
import sys

from ldapom import compat

if sys.version_info[0] >= 3: # Python 3
    unicode = str

OID_REGEX = re.compile(r"\( ([0-9\.]+) .* \)")
SINGLE_VALUE_REGEX = re.compile(r"\(.* SINGLE-VALUE .*\)")
ONE_NAME_REGEX = re.compile(r"\(.* NAME \'([^']+)\' .*\)")
MULTIPLE_NAMES_REGEX = re.compile(r"\(.* NAME \( ([^)]+) \) .*\)")
SYNTAX_REGEX = re.compile(r"\(.* SYNTAX ([0-9\.]+)({\d+})? .*\)")
DESC_REGEX = re.compile(r"\(.* DESC \'([^']+)\' .*\)")
SUP_REGEX = re.compile(r"\(.* SUP (\w+) .*\)")


class LDAPAttributeBase(compat.UnicodeMixin, object):
    """Holds an LDAP attribute and its values."""

    def __init__(self, name):
        """Creates a new attribute.

        :param name: The name for the attribute.
        :param name: str
        """
        self.name = name
        self._values = set()

    def __unicode__(self):
        if len(self._values) == 1:
            values_string = unicode(next(iter(self._values)))
        else:
            values_string = unicode(", ".join(map(unicode, self._values)))
        return "{name}: {values}".format(name=self.name, values=values_string)

    def __repr__(self):
        return "<LDAPAttribute " + self.name + ">"

    def __eq__(self, other):
        return self.name == other.name and self._values == other._values

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.name, frozenset(self._values)))

    def __deepcopy__(self, memo):
        c = type(self)(self.name)
        c._values = copy.deepcopy(self._values)
        return c


class SingleValueAttributeMixin(object):
    single_value = True
    multi_value = False

    def _get_value(self):
        if len(self._values) == 0:
            return None
        else:
            return next(iter(self._values))

    def _set_value(self, value):
        if value is None:
            self._values = set()
        else:
            self._values = set([value])

    value = property(_get_value, _set_value)


class MultiValueAttributeMixin(object):
    single_value = False
    multi_value = True

    def _get_values(self):
        return self._values

    def _set_values(self, values):
        self._values = set(values)

    values = property(_get_values, _set_values)


class UnicodeAttributeMixin(object):
    def _set_ldap_values(self, values):
        self._values = set([compat._decode_utf8(v) for v in values])

    def _get_ldap_values(self):
        return set([compat._encode_utf8(v) for v in self._values])


class BooleanAttributeMixin(object):
    def _set_ldap_values(self, values):
        self._values = set([(compat._decode_utf8(v) == "TRUE") for v in values])

    def _get_ldap_values(self):
        string_values = ["TRUE" if v else "FALSE" for v in self._values]
        return set([compat._encode_utf8(v) for v in string_values])


class IntegerAttributeMixin(object):
    def _set_ldap_values(self, values):
        self._values = set([int(compat._decode_utf8(v)) for v in values])

    def _get_ldap_values(self):
        return set([compat._encode_utf8(unicode(v)) for v in self._values])


class BytesAttributeMixin(object):
    def _set_ldap_values(self, values):
        self._values = set(values)

    def _get_ldap_values(self):
        return set(self._values)


ATTRIBUTE_SYNTAX_TO_TYPE_MIXIN = {
        '1.3.6.1.1.1.0.0':               BytesAttributeMixin,     # RFC2307 NIS Netgroup Triple
        '1.3.6.1.1.1.0.1':               UnicodeAttributeMixin, # RFC2307 Boot Parameter
        '1.3.6.1.1.16.1':                UnicodeAttributeMixin, # UUID
        '1.3.6.1.4.1.1466.115.121.1.3':  UnicodeAttributeMixin, # Attribute Type Description
        '1.3.6.1.4.1.1466.115.121.1.4':  BytesAttributeMixin,     # Audio
        '1.3.6.1.4.1.1466.115.121.1.5':  BytesAttributeMixin,     # Binary
        '1.3.6.1.4.1.1466.115.121.1.6':  BytesAttributeMixin,     # Bit String
        '1.3.6.1.4.1.1466.115.121.1.7':  BooleanAttributeMixin,    # Boolean
        '1.3.6.1.4.1.1466.115.121.1.8':  BytesAttributeMixin,     # Certificate
        '1.3.6.1.4.1.1466.115.121.1.9':  BytesAttributeMixin,     # Certificate List
        '1.3.6.1.4.1.1466.115.121.1.10': BytesAttributeMixin,     # Certificate Pair
        '1.3.6.1.4.1.1466.115.121.1.11': UnicodeAttributeMixin, # CountryString
        '1.3.6.1.4.1.1466.115.121.1.12': UnicodeAttributeMixin, # Distinguished Name
        '1.3.6.1.4.1.1466.115.121.1.13': UnicodeAttributeMixin, # Data Quality Syntax
        '1.3.6.1.4.1.1466.115.121.1.14': UnicodeAttributeMixin, # Delivery Method
        '1.3.6.1.4.1.1466.115.121.1.15': UnicodeAttributeMixin, # DirectoryString
        '1.3.6.1.4.1.1466.115.121.1.19': UnicodeAttributeMixin, # DSA Quality Syntax
        '1.3.6.1.4.1.1466.115.121.1.21': UnicodeAttributeMixin, # Enhanced Guide
        '1.3.6.1.4.1.1466.115.121.1.22': UnicodeAttributeMixin, # Facsimile Telephone Number
        '1.3.6.1.4.1.1466.115.121.1.23': BytesAttributeMixin,     # Fax Image Syntax
        '1.3.6.1.4.1.1466.115.121.1.24': UnicodeAttributeMixin, # GeneralizedTime
        '1.3.6.1.4.1.1466.115.121.1.25': UnicodeAttributeMixin, # Guide (Obsolete)
        '1.3.6.1.4.1.1466.115.121.1.26': UnicodeAttributeMixin, # IA5String
        '1.3.6.1.4.1.1466.115.121.1.27': IntegerAttributeMixin,     # Integer
        '1.3.6.1.4.1.1466.115.121.1.28': BytesAttributeMixin,     # JPEG
        '1.3.6.1.4.1.1466.115.121.1.30': UnicodeAttributeMixin, # Matching Rule Description syntax
        '1.3.6.1.4.1.1466.115.121.1.31': UnicodeAttributeMixin, # Matching Rule Use Description syntax
        '1.3.6.1.4.1.1466.115.121.1.34': UnicodeAttributeMixin, # Name And Optional UID
        '1.3.6.1.4.1.1466.115.121.1.36': UnicodeAttributeMixin, # NumericString
        '1.3.6.1.4.1.1466.115.121.1.37': UnicodeAttributeMixin, # Object Class Description syntax
        '1.3.6.1.4.1.1466.115.121.1.38': UnicodeAttributeMixin, # OID
        '1.3.6.1.4.1.1466.115.121.1.39': UnicodeAttributeMixin, # Other Mailbox
        '1.3.6.1.4.1.1466.115.121.1.40': BytesAttributeMixin,     # OctetString
        '1.3.6.1.4.1.1466.115.121.1.41': UnicodeAttributeMixin, # PostalAddress
        '1.3.6.1.4.1.1466.115.121.1.42': UnicodeAttributeMixin, # protocolInformation
        '1.3.6.1.4.1.1466.115.121.1.43': UnicodeAttributeMixin, # Presentation Address syntax
        '1.3.6.1.4.1.1466.115.121.1.44': UnicodeAttributeMixin, # PrintableString
        '1.3.6.1.4.1.1466.115.121.1.49': BytesAttributeMixin,     # Supported Algorithm
        '1.3.6.1.4.1.1466.115.121.1.50': UnicodeAttributeMixin, # TelephoneNumber
        '1.3.6.1.4.1.1466.115.121.1.51': UnicodeAttributeMixin, # Teletex Terminal Identifier
        '1.3.6.1.4.1.1466.115.121.1.52': UnicodeAttributeMixin, # Telex Number
        '1.3.6.1.4.1.1466.115.121.1.54': UnicodeAttributeMixin, # LDAP Syntax Description
        '1.3.6.1.4.1.4203.666.2.7':      BytesAttributeMixin,     # OpenLDAP authz
        # Experimental syntax used by OpenLDAP, see ITS#5573
        '1.3.6.1.4.1.4203.666.11.2.1':   UnicodeAttributeMixin, # OpenLDAP CSN
        }


DEFAULT_ATTRIBUTE_TYPES = [
    # Attribute type used by OpenLDAP, but not standardized or advertized in subschema.
    # See ITS#5573
    "( 1.3.6.1.4.1.4203.666.1.7 NAME 'entryCSN' "
    "DESC 'change sequence number of the entry content' "
    "EQUALITY CSNMatch "
    "ORDERING CSNOrderingMatch "
    "SYNTAX 1.3.6.1.4.1.4203.666.11.2.1{64} "
    "SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )"]


def build_attribute_types(type_definitions):
    """Build attribute types from type definitions.

    :param type_definitions: A list of LDAP attribute type definitions.
    :type type_definition: list of str
    :rtype: Name-keyed dict of attribute types.
    """
    attribute_type_dicts = []
    for type_definition in type_definitions:
        type_dict = {}

        sup_match = SUP_REGEX.match(type_definition)
        if sup_match:
            type_dict["sup"] = sup_match.group(1)

        desc_match = DESC_REGEX.match(type_definition)
        if desc_match:
            type_dict["desc"] = desc_match.group(1)

        oid_match = OID_REGEX.match(type_definition)
        if oid_match:
            type_dict["oid"] = oid_match.group(1)

        single_value_match = SINGLE_VALUE_REGEX.match(type_definition)
        if single_value_match:
            type_dict["single_value"] = bool(single_value_match)

        one_name_match = ONE_NAME_REGEX.match(type_definition)
        if one_name_match:
            names = [one_name_match.group(1)]
        else:
            names_string = MULTIPLE_NAMES_REGEX.match(type_definition).group(1)
            names = [n.strip("'") for n in names_string.split(" ")]
        type_dict["names"] = names

        syntax_match = SYNTAX_REGEX.match(type_definition)
        if syntax_match:
            type_dict["syntax"] = syntax_match.group(1)

        attribute_type_dicts.append(type_dict)

    type_dicts_by_name = {}
    for type_dict in attribute_type_dicts:
        for name in type_dict["names"]:
            type_dicts_by_name[name] = type_dict

    # Resolve inheritance for attribute types
    resolved_type_dicts = []
    for type_dict in attribute_type_dicts:
        # Build a list of ancestors, starting with the leaf
        ancestors = []
        current_ancestor = type_dict
        while current_ancestor is not None:
            ancestors.append(current_ancestor)
            current_ancestor = type_dicts_by_name.get(
                    current_ancestor.get("sup", None), None)
        # Reverse ancestors, so the root comes first
        ancestors.reverse()
        resolved_type_dict = {}
        for ancestor in ancestors:
            resolved_type_dict.update(ancestor)

        resolved_type_dicts.append(resolved_type_dict)

    # Build the types for each of the type definitions
    attribute_types_by_name = {}
    for type_dict in resolved_type_dicts:
        base_classes = []
        if type_dict.get("single_value", False):
            base_classes.append(SingleValueAttributeMixin)
        else:
            base_classes.append(MultiValueAttributeMixin)

        type_mixin = ATTRIBUTE_SYNTAX_TO_TYPE_MIXIN[type_dict["syntax"]]
        base_classes.append(type_mixin)

        base_classes.append(LDAPAttributeBase)
        if sys.version_info[0] >= 3: # Python 3
            attribute_type = type("LDAPAttribute", tuple(base_classes), {})
        else:
            attribute_type = type(bytes("LDAPAttribute"),
                    tuple(base_classes), {})
        for name in type_dict["names"]:
            attribute_types_by_name[name] = attribute_type

    return attribute_types_by_name
